import base64
import logging
import os
import re

from marshmallow.exceptions import ValidationError
from mutagen import MutagenError
from mutagen.mp3 import EasyMP3
from mutagen.easymp4 import EasyMP4
from mutagen.flac import FLAC
from mutagen.apev2 import APEv2

from feeluown.utils.utils import elfhash
from feeluown.models import AlbumType

from .lans_helpers import core_lans
from .provider import read_audio_cover, Media, reverse, MediaType

from .schemas import (
    EasyMP3MetadataSongSchema,
    FLACMetadataSongSchema,
    APEMetadataSongSchema,
)
from .models import (
    LSongModel,
    LAlbumModel,
    LArtistModel,
)

logger = logging.getLogger(__name__)


def gen_id(s):
    return str(elfhash(base64.b64encode(bytes(s, 'utf-8'))))


def create_artist(identifier, name):
    return LArtistModel(identifier=identifier,
                        name=name,
                        songs=[],
                        albums=[],
                        contributed_albums=[],
                        desc='',
                        cover='',)


def create_album(identifier, name, cover):
    """create album model with album name
    """
    album = LAlbumModel(identifier=identifier,
                        name=name,
                        songs=[],
                        artists=[],
                        desc='',
                        cover=cover,)
    # guess album type by its name
    #
    # album name which contains following string are `Single`
    #   1. ' - Single'  6+3=9
    #   2. '(single)'   6+2=8
    #   3. '（single）'  6+2=8
    if 'single' in name[-9:].lower():
        album.type = AlbumType.single
    if 'ep' in name[-5:].lower():
        album.type = AlbumType.ep
    return album


def add_song(fpath, g_songs, g_artists, g_albums, lans='auto', delimiter='', expand_artist_songs=False):
    """
    parse music file metadata with Easymp3 and return a song
    model.
    """
    try:
        if fpath.endswith('mp3') or fpath.endswith('ogg') or fpath.endswith('wma'):
            metadata = EasyMP3(fpath)
        elif fpath.endswith('m4a') or fpath.endswith('m4v') or fpath.endswith('mp4'):
            metadata = EasyMP4(fpath)
        elif fpath.endswith('flac'):
            metadata = FLAC(fpath)
        elif fpath.endswith('ape'):
            metadata = APEv2(fpath)
        elif fpath.endswith('wav'):
            metadata = dict()
    except MutagenError as e:
        logger.warning(
            'Mutagen parse metadata failed, ignore.\n'
            'file: {}, exception: {}'.format(fpath, str(e)))
        return None

    metadata_dict = dict(metadata)
    for key in metadata.keys():
        metadata_dict[key] = core_lans(metadata_dict[key][0], lans)
    if 'title' not in metadata_dict:
        title = os.path.split(fpath)[-1].split('.')[0]
        metadata_dict['title'] = core_lans(title, lans)
    metadata_dict.update(dict(
        url=fpath,
        duration=metadata.info.length * 1000  # milesecond
    ))

    try:
        if fpath.endswith('flac'):
            data = FLACMetadataSongSchema().load(metadata_dict)
        elif fpath.endswith('ape'):
            data = APEMetadataSongSchema().load(metadata_dict)
        else:
            data = EasyMP3MetadataSongSchema().load(metadata_dict)
    except ValidationError:
        logger.exception('解析音乐文件({}) 元数据失败'.format(fpath))
        return

    # NOTE: use {title}-{artists_name}-{album_name} as song identifier
    title = data['title']
    album_name = data['album_name']
    artist_name_list = [
        name.strip()
        for name in re.split(r'[,&]', data['artists_name'])]
    artists_name = ','.join(artist_name_list)
    duration = data['duration']
    album_artist_name = data['album_artist_name']

    # 生成 song model
    # 用来生成 id 的字符串应该尽量减少无用信息，这样或许能减少 id 冲突概率
    # 加入分隔符'-'在一定概率上更能确保不发生哈希值重复
    song_id_str = delimiter.join([title, artists_name, album_name, str(int(duration))])
    song_id = gen_id(song_id_str)
    if song_id not in g_songs:
        # 剩下 album, lyric 三个字段没有初始化
        song = LSongModel(identifier=song_id,
                          artists=[],
                          title=title,
                          url=fpath,
                          duration=duration,
                          comments=[],
                          # 下面这些字段不向外暴露
                          genre=data['genre'],
                          cover=data['cover'],
                          date=data['date'],
                          desc=data['desc'],
                          disc=data['disc'],
                          track=data['track'])
        g_songs[song_id] = song
    else:
        song = g_songs[song_id]
        logger.warning('Duplicate song: %s %s', song.url, fpath)
        return

    # 生成 album artist model
    album_artist_id = gen_id(album_artist_name)
    if album_artist_id not in g_artists:
        album_artist = create_artist(album_artist_id, album_artist_name)
        g_artists[album_artist_id] = album_artist
    else:
        album_artist = g_artists[album_artist_id]

    # 生成 album model
    album_id_str = delimiter.join([album_name, album_artist_name])
    album_id = gen_id(album_id_str)
    # cover_data, cover_fmt = read_audio_cover(fpath)
    # if cover_data is None:
    #     cover = None
    # else:
    #     cover = Media(reverse(song, '/cover/data'), type_=MediaType.image)
    if album_id not in g_albums:
        album = create_album(album_id, album_name, None)
        g_albums[album_id] = album
    else:
        album = g_albums[album_id]

    # 处理专辑的歌手信息和歌曲信息，专辑歌手的专辑列表信息
    if album not in album_artist.albums:
        album_artist.albums.append(album)
    if album_artist not in album.artists:
        album.artists.append(album_artist)
    if song not in album.songs:
        album.songs.append(song)

    # 处理歌曲的歌手和专辑信息，以及歌手的歌曲列表和参与作品
    song.album = album
    for artist_name in artist_name_list:
        artist_id = gen_id(artist_name)
        if artist_id in g_artists:
            artist = g_artists[artist_id]
        else:
            artist = create_artist(identifier=artist_id, name=artist_name)
            g_artists[artist_id] = artist
        if artist not in song.artists:
            song.artists.append(artist)
        if song not in artist.songs:
            artist.songs.append(song)

        # 处理歌曲歌手的参与作品信息(不与前面的重复)
        if album not in artist.albums and album not in artist.contributed_albums:
            artist.contributed_albums.append(album)

    # 处理专辑歌手的歌曲信息: 有些作词人出合辑很少出现在歌曲歌手里(可选)
    if expand_artist_songs and song not in album_artist.songs:
        album_artist.songs.append(song)


class DB:
    """
    DB manages a fileset and their corresponding models

    the data structure in db file::

        {
          "files": [],
          "songs": []
        }
    """
    def __init__(self, fpath):
        """
        :param filepath: database file path
        """

        self._dirty = False  # whether changes are flushed
        self._fileset = set()  # media files

        self._songs = []
        self._artists = []
        self._albums = []

    def flush(self):
        """flush the changes into db file"""

    def list_models(self):
        """list all models in database"""
        pass

    def add(self, fpath):
        """add media file to database"""
        add_song(fpath, self._songs, self._artists, self._albums)

    def remove(self, fpath):
        """add media file to database"""



# -*- coding: utf-8 -*-
from marshmallow import Schema, fields, post_load, EXCLUDE


DEFAULT_TITLE = DEFAULT_ARTIST_NAME = DEFAULT_ALBUM_NAME = 'Unknown'


class EasyMP3MetadataSongSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    """EasyMP3 metadata"""
    url = fields.Str(required=True)
    duration = fields.Float(required=True)
    title = fields.Str(missing=DEFAULT_TITLE)
    artists_name = fields.Str(data_key='artist', missing=DEFAULT_ARTIST_NAME)
    album_name = fields.Str(data_key='album', missing=DEFAULT_ALBUM_NAME)
    album_artist_name = fields.Str(data_key='albumartist', missing=DEFAULT_ARTIST_NAME)
    track = fields.Str(data_key='tracknumber', missing='1/1')
    disc = fields.Str(data_key='discnumber', missing='1/1')
    date = fields.Str(missing='')
    genre = fields.Str(missing='')

    # 本地歌曲目前都不支持描述和封面，将它们设置为空字符串
    desc = fields.Str(missing='')
    cover = fields.Str(missing='')


class FLACMetadataSongSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    """FLAC metadata"""
    url = fields.Str(required=True)
    duration = fields.Float(required=True)
    title = fields.Str(missing=DEFAULT_TITLE)
    artists_name = fields.Str(data_key='artist', missing=DEFAULT_ARTIST_NAME)
    album_name = fields.Str(data_key='album', missing=DEFAULT_ALBUM_NAME)
    album_artist_name = fields.Str(data_key='albumartist', missing=DEFAULT_ARTIST_NAME)
    track = fields.Str(data_key='tracknumber', missing='1')
    disc = fields.Str(data_key='discnumber', missing='1')
    tracktotal = fields.Str(data_key='tracktotal', missing='1')
    disctotal = fields.Str(data_key='disctotal', missing='1')
    date = fields.Str(missing='')
    genre = fields.Str(missing='')

    # 本地歌曲目前都不支持描述和封面，将它们设置为空字符串
    desc = fields.Str(missing='')
    cover = fields.Str(missing='')

    @post_load
    def create_model(self, data, **kwargs):
        data['track'] = '{}/{}'.format(data['track'], data.pop('tracktotal'))
        data['disc'] = '{}/{}'.format(data['disc'], data.pop('disctotal'))
        return data


class APEMetadataSongSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    """APE metadata"""
    url = fields.Str(required=True)
    duration = fields.Float(required=True)
    title = fields.Str(missing=DEFAULT_TITLE)
    artists_name = fields.Str(data_key='artist', missing=DEFAULT_ARTIST_NAME)
    album_name = fields.Str(data_key='album', missing=DEFAULT_ALBUM_NAME)
    album_artist_name = fields.Str(data_key='albumartist', missing=DEFAULT_ARTIST_NAME)
    track = fields.Str(data_key='track', missing='1/1')
    disc = fields.Str(data_key='discnumber', missing='1/1')
    date = fields.Str(missing='')
    genre = fields.Str(missing='')

    # 本地歌曲目前都不支持描述和封面，将它们设置为空字符串
    desc = fields.Str(missing='')
    cover = fields.Str(missing='')



import logging


logger = logging.getLogger(__name__)


def core_lans(str, lans):
    try:
        assert lans in ['auto', 'cn', 'tc']
        if lans in ['cn', 'tc']:
            import inlp.convert.chinese as cv
            return cv.t2s(str) if lans == 'cn' else cv.s2t(str)
        else:
            return str
    except Exception as e:
        logger.warning(e)
        return str



# -*- coding: utf-8 -*-
# pylint: disable=wrong-import-position
"""
TODO: 这个模块中目前逻辑非常多，包括音乐目录扫描、音乐库的构建等小部分，
这些小部分理论都可以从中拆除。
"""

import logging
import os

from fuzzywuzzy import process
from feeluown.library import AbstractProvider, ProviderV2, ProviderFlags as PF

from feeluown.utils import aio
from feeluown.utils.utils import log_exectime
from feeluown.media import Media, MediaType
from feeluown.models import reverse
from feeluown.models import ModelType, SearchType

from .utils import read_audio_cover

logger = logging.getLogger(__name__)
SOURCE = 'local'


def scan_directory(directory, exts=None, depth=2):
    exts = exts or ['mp3', 'fuo']
    if depth < 0:
        return []

    media_files = []
    if not os.path.exists(directory):
        return []
    for path in os.listdir(directory):
        path = os.path.join(directory, path)
        if os.path.isdir(path):
            files = scan_directory(path, exts, depth - 1)
            media_files.extend(files)
        elif os.path.isfile(path):
            if path.split('.')[-1] in exts:
                media_files.append(path)
    return media_files


class Library:
    def __init__(self):
        self._songs = {}
        self._albums = {}
        self._artists = {}

    def list_songs(self):
        return list(self._songs.values())

    def list_albums(self):
        return list(self._albums.values())

    def list_artists(self):
        return list(self._artists.values())

    def get_song(self, identifier):
        return self._songs.get(identifier)

    def get_album(self, identifier):
        return self._albums.get(identifier)

    def get_artist(self, identifier):
        return self._artists.get(identifier)

    @log_exectime
    def scan(self, config, paths, depth, exts):
        """scan media files in all paths
        """
        media_files = []
        logger.info('start scanning...')
        for directory in paths:
            logger.debug('正在扫描目录(%s)...', directory)
            media_files.extend(scan_directory(directory, exts, depth))
        logger.info(f'scanning finished, {len(media_files)} files in total')

        for fpath in media_files:
            add_song(fpath, self._songs, self._artists, self._albums,
                     config.CORE_LANGUAGE,
                     config.IDENTIFIER_DELIMITER,
                     config.EXPAND_ARTIST_SONGS)
        logger.info('录入本地音乐库完毕')

    def after_scan(self):
        """
        歌曲扫描完成后，对信息进行一些加工，比如
        1. 给专辑歌曲排序
        2. 给专辑和歌手加封面
        """
        def sort_album_func(album):
            if album.songs:
                return (album.songs[0].date is not None, album.songs[0].date)
            return (False, '0')

        for album in self._albums.values():
            try:
                album.songs.sort(key=lambda x: (int(x.disc.split('/')[0]),
                                                int(x.track.split('/')[0])))
                if album.name != 'Unknown':
                    cover_data, _ = read_audio_cover(album.songs[0].url)
                    if cover_data:
                        cover = Media(reverse(album.songs[0], '/cover/data'),
                                      type_=MediaType.image)
                    else:
                        cover = None
                    album.cover = cover
            except:  # noqa
                logger.exception('Sort album songs failed.')

        for artist in self._artists.values():
            if artist.albums:
                artist.albums.sort(key=sort_album_func, reverse=True)
                artist.cover = artist.albums[0].cover
            if artist.contributed_albums:
                artist.contributed_albums.sort(key=sort_album_func, reverse=True)
            if artist.songs:
                # sort artist songs
                artist.songs.sort(key=lambda x: x.title)
                # use song cover as artist cover
                # https://github.com/feeluown/feeluown-local/pull/3/files#r362126996
                songs_with_unknown_album = [song for song in artist.songs
                                            if song.album_name == 'Unknown']
                for song in sorted(songs_with_unknown_album,
                                   key=lambda x: (x.date is not None, x.date),
                                   reverse=True):
                    if read_audio_cover(song.url)[0]:
                        artist.cover = Media(reverse(song, '/cover/data'),
                                             type_=MediaType.image)
                        break


class LocalProvider(AbstractProvider, ProviderV2):
    class meta:
        identifier = SOURCE
        name = '本地音乐'
        flags = {
            ModelType.song: (PF.lyric),
        }

    def __init__(self):
        super().__init__()

        self._app = None
        self.library = Library()

    def initialize(self, app):
        self._app = app

    def scan(self, config, paths, depth=3):
        exts = config.MUSIC_FORMATS
        self.library.scan(config, paths, depth, exts)
        self.library.after_scan()

    @property
    def identifier(self):
        return SOURCE

    @property
    def name(self):
        return '本地音乐'

    @property
    def songs(self):
        return self.library.list_songs()

    @property
    def albums(self):
        return self.library.list_albums()

    @property
    def artists(self):
        return self.library.list_artists()

    @log_exectime
    def search(self, keyword, **kwargs):
        limit = kwargs.get('limit', 10)
        repr_song_map = dict()
        for song in self.songs:
            key = song.title + ' ' + song.artists_name
            repr_song_map[key] = song
        choices = repr_song_map.keys()
        if choices:
            result = process.extract(keyword, choices, limit=limit)
        else:
            result = []
        result_songs = []
        for each, score in result:
            # if score > 80, keyword is almost included in song key
            if score > 80:
                result_songs.append(repr_song_map[each])
        return LSearchModel(q=keyword, songs=result_songs)

    def song_get_lyric(self, song):
        # 歌词获取报错的 workaround
        if self._app is None:
            return None
        provider = self._app.library.get('qqmusic')
        if provider is None:
            return None
        result = provider.search(f'{song.title} {song.artists_name}', type_=SearchType.so)
        songs = result.songs
        if len(songs) < 1:
            return None
        return provider.song_get_lyric(songs[0]) or songs[0].lyric


provider = LocalProvider()


from .db import add_song  # noqa
from .models import LSearchModel  # noqa



import logging

from feeluown.models import (
    BaseModel,
    SongModel,
    AlbumModel,
    ArtistModel,
    SearchModel,
)
from feeluown.utils.reader import wrap

from .provider import provider
from .utils import read_audio_cover

logger = logging.getLogger(__name__)


class LBaseModel(BaseModel):

    class Meta:
        allow_get = True
        provider = provider


class LSongModel(SongModel, LBaseModel):

    class Meta:
        fields = ('disc', 'genre', 'date', 'track', 'cover', 'desc')
        fields_no_get = ('lyric', )
        paths = [
            '/cover/data',
        ]

    @classmethod
    def get(cls, identifier):
        return cls.meta.provider.library.get_song(identifier)

    @classmethod
    def list(cls, identifier_list):
        return map(cls.meta.provider.library._songs.get, identifier_list)

    def resolve__cover_data(self, **kwargs):
        return read_audio_cover(self.url)[0]


class LAlbumModel(AlbumModel, LBaseModel):

    @classmethod
    def get(cls, identifier):
        return cls.meta.provider.library.get_album(identifier)


class LArtistModel(ArtistModel, LBaseModel):

    class Meta:
        fields = ('contributed_albums', )
        allow_create_albums_g = True

    @classmethod
    def get(cls, identifier):
        return cls.meta.provider.library.get_artist(identifier)

    def create_albums_g(self):
        return wrap(self.albums)

    def create_contributed_albums_g(self):
        return wrap(self.contributed_albums)


class LSearchModel(SearchModel, LBaseModel):
    pass



from mutagen.id3 import _specs


def patch_mutagen():
    old_decode_terminated = _specs.decode_terminated

    def decode_terminated(data, encoding, strict=True):
        """
        国内部分歌曲的 id3 信息是使用 GBK 编码，但是它的元信息显示自己是
        latin1 编码，所以 mutagen 解析的时候会返回不正确的结果。

        这里，我们没有选择重新实现原函数，而是对原函数的返回进行加工，
        这样可以比较好地确保我们的实现不会影响到 mutagen 的其它逻辑。
        但这样实现有一个问题，原函数已经进行过 decode 和 encode，我们这里又重复
        encode 和 decode，理论上会带来一定的性能损耗，欢迎有兴趣的同鞋进行优化。

        ps: 根据 id3 2.3/2.4 的规范，gbk 编码 *似乎* 本来就是非法的，
        规范上说只支持这四种： latin1, utf16, utf_16_be, utf8，
        mutagen 确实也是按照 spec 来实现的，但现实很残酷。
        """
        orig_value, data = old_decode_terminated(data, encoding, strict)
        value = orig_value
        if encoding == 'latin1':
            value = orig_value.encode('latin1')
            # 参考 deepin-music-player 项目 easymp3.py 实现
            for codec in ['gbk', 'big5', 'utf-8', 'iso-8559-15', 'cp1251']:
                try:
                    value = value.decode(codec)
                except (UnicodeDecodeError, LookupError):
                    # decode failed or codec not found
                    pass
                else:
                    break
            else:
                value = orig_value
        return value, data

    _specs.decode_terminated = decode_terminated



# -*- coding: utf-8 -*-
import logging
from functools import partial

from feeluown.utils import aio

from .patch import patch_mutagen
patch_mutagen()

from .consts import DEFAULT_MUSIC_FOLDER, DEFAULT_MUSIC_EXTS
from .provider import provider


__alias__ = '本地音乐'
__feeluown_version__ = '1.1.0'
__version__ = '0.1a0'
__desc__ = '本地音乐'

logger = logging.getLogger(__name__)


def init_config(config):
    config.deffield('MUSIC_FOLDERS', type_=list, default=[DEFAULT_MUSIC_FOLDER], desc='')
    config.deffield('MUSIC_FORMATS', type_=list, default=DEFAULT_MUSIC_EXTS, desc='')
    config.deffield('CORE_LANGUAGE', type_=str, default='auto', desc='')
    config.deffield('IDENTIFIER_DELIMITER', type_=str, default='', desc='')
    config.deffield('EXPAND_ARTIST_SONGS', type_=bool, default=False, desc='')


async def autoload(app):
    await aio.run_fn(provider.scan,
                     app.config.fuo_local,
                     app.config.fuo_local.MUSIC_FOLDERS)

    app.show_msg('本地音乐扫描完毕')
    if app.mode & app.GuiMode:
        app.coll_uimgr.refresh()


def show_provider(req):
    from .ui import LibraryRenderer
    if hasattr(req, 'ctx'):
        app = req.ctx['app']
    else:
        app = req  # 兼容老版本
    app.pl_uimgr.clear()
    # app.playlists.add(provider.playlists)

    app.ui.left_panel.my_music_con.hide()
    app.ui.left_panel.playlists_con.hide()

    aio.run_afn(
        app.ui.table_container.set_renderer,
        LibraryRenderer(provider.songs, provider.albums, provider.artists))


def enable(app):
    logger.info('Register provider: %s', provider)
    app.library.register(provider)
    provider.initialize(app)

    app.initialized.connect(lambda *args: aio.create_task(autoload(*args)),
                            weak=False, aioqueue=False)
    if app.mode & app.GuiMode:
        app.browser.route('/local')(show_provider)
        pm = app.pvd_uimgr.create_item(
            name=provider.identifier,
            text='本地音乐',
            symbol='♪ ',
            desc='点击显示所有本地音乐',
        )
        pm.clicked.connect(partial(app.browser.goto, uri='/local'), weak=False)
        app.pvd_uimgr.add_item(pm)


def disable(app):
    logger.info('唔，不要禁用我')



"""
scanner scan a directory and build a
"""

from enum import IntEnum

KNOWN_EXTS = ('mp3', 'ogg', 'wma', 'm4a', 'm4v', 'flac')


class FileEventType(IntEnum):
    deleted = 0
    added = 1
    changed = 2


class FileEvent:
    def __init__(self, filepath, type_):
        pass


class Scanner:
    def __init__(self, paths, exts=None, depth=2):
        """

        :return: {filepath: checksum, ...}
        """
        self._exts = exts or KNOWN_EXTS
        # common directory structure are as following
        # * Music/<provider>/<artist>/<album>/<title>.<ext>
        # * Music/<provider>/<album>/<title - artist>.<ext>
        # * Music/<provider>/<title - artist>.<ext>
        # therefore, we set max depth to 3, which should be large enough
        self._depth = min(3, depth)
        self._paths = paths
